@@ -9,7 +9,10 @@ use std::{
 use parking_lot::Mutex;
 
 use super::{FileManager, FileOp, ManagedFile, OpenableFile};
-use crate::{error::Error, io::PathIds};
+use crate::{
+    error::Error,
+    io::{File as _, ManagedFileOpener, OperableFile, PathIds},
+};
 
 /// An open file that uses [`std::fs`].
 #[derive(Debug)]
@@ -21,6 +24,9 @@ pub struct StdFile {
 
 impl ManagedFile for StdFile {
     type Manager = StdFileManager;
+}
+
+impl super::File for StdFile {
     fn id(&self) -> Option<u64> {
         self.id
     }
@@ -29,35 +35,6 @@ impl ManagedFile for StdFile {
         &self.path
     }
 
-    fn open_for_read(
-        path: impl AsRef<std::path::Path> + Send,
-        id: Option<u64>,
-    ) -> Result<Self, Error> {
-        let path = path.as_ref();
-        Ok(Self {
-            file: File::open(path)?,
-            path: path.to_path_buf(),
-            id,
-        })
-    }
-
-    fn open_for_append(
-        path: impl AsRef<std::path::Path> + Send,
-        id: Option<u64>,
-    ) -> Result<Self, Error> {
-        let path = path.as_ref();
-        Ok(Self {
-            file: OpenOptions::new()
-                .write(true)
-                .append(true)
-                .read(true)
-                .create(true)
-                .open(path)?,
-            path: path.to_path_buf(),
-            id,
-        })
-    }
-
     fn length(&self) -> Result<u64, Error> {
         let metadata = self.file.metadata()?;
         Ok(metadata.len())
@@ -69,6 +46,42 @@ impl ManagedFile for StdFile {
     }
 }
 
+/// A [`ManagedFileOpener`] implementation that produces [`StdFile`]s.
+pub struct StdFileOpener;
+
+impl ManagedFileOpener<StdFile> for StdFileOpener {
+    fn open_for_read(
+        &self,
+        path: impl AsRef<std::path::Path> + Send,
+        id: Option<u64>,
+    ) -> Result<StdFile, Error> {
+        let path = path.as_ref();
+        Ok(StdFile {
+            file: File::open(path)?,
+            path: path.to_path_buf(),
+            id,
+        })
+    }
+
+    fn open_for_append(
+        &self,
+        path: impl AsRef<std::path::Path> + Send,
+        id: Option<u64>,
+    ) -> Result<StdFile, Error> {
+        let path = path.as_ref();
+        Ok(StdFile {
+            file: OpenOptions::new()
+                .write(true)
+                .append(true)
+                .read(true)
+                .create(true)
+                .open(path)?,
+            path: path.to_path_buf(),
+            id,
+        })
+    }
+}
+
 impl Seek for StdFile {
     #[cfg_attr(feature = "tracing", tracing::instrument(skip(self,)))]
     fn seek(&mut self, pos: SeekFrom) -> std::io::Result<u64> {
@@ -115,7 +128,7 @@ impl FileManager for StdFileManager {
     type FileHandle = OpenStdFile;
     fn append(&self, path: impl AsRef<Path>) -> Result<Self::FileHandle, Error> {
         let path = path.as_ref();
-        let file_id = self.file_ids.file_id_for_path(path);
+        let file_id = self.file_ids.file_id_for_path(path, true).unwrap();
         let mut open_files = self.open_files.lock();
         if let Some(open_file) = open_files.get_mut(&file_id) {
             let mut file = FileSlot::Taken;
@@ -146,7 +159,7 @@ impl FileManager for StdFileManager {
                 manager: Some(self.clone()),
             })
         } else {
-            let file = StdFile::open_for_append(path, Some(file_id))?;
+            let file = self.open_for_append(path, Some(file_id))?;
             open_files.insert(file_id, FileSlot::Taken);
             Ok(OpenStdFile {
                 file: Some(file),
@@ -158,7 +171,7 @@ impl FileManager for StdFileManager {
 
     fn read(&self, path: impl AsRef<Path>) -> Result<Self::FileHandle, Error> {
         let path = path.as_ref();
-        let file_id = self.file_ids.file_id_for_path(path);
+        let file_id = self.file_ids.file_id_for_path(path, true).unwrap();
 
         let mut reader_files = self.reader_files.lock();
         let files = reader_files.entry(file_id).or_default();
@@ -171,7 +184,7 @@ impl FileManager for StdFileManager {
             });
         }
 
-        let file = StdFile::open_for_read(path, Some(file_id))?;
+        let file = StdFileOpener.open_for_read(path, Some(file_id))?;
         Ok(OpenStdFile {
             file: Some(file),
             manager: Some(self.clone()),
@@ -223,6 +236,35 @@ impl FileManager for StdFileManager {
             publish_callback(result.new_id);
         }
     }
+
+    fn exists(&self, path: impl AsRef<std::path::Path>) -> Result<bool, crate::Error> {
+        Ok(path.as_ref().exists())
+    }
+
+    fn file_length(&self, path: impl AsRef<Path>) -> Result<u64, Error> {
+        path.as_ref()
+            .metadata()
+            .map_err(Error::from)
+            .map(|metadata| metadata.len())
+    }
+}
+
+impl ManagedFileOpener<StdFile> for StdFileManager {
+    fn open_for_read(
+        &self,
+        path: impl AsRef<Path> + Send,
+        id: Option<u64>,
+    ) -> Result<StdFile, Error> {
+        StdFileOpener.open_for_read(path, id)
+    }
+
+    fn open_for_append(
+        &self,
+        path: impl AsRef<Path> + Send,
+        id: Option<u64>,
+    ) -> Result<StdFile, Error> {
+        StdFileOpener.open_for_append(path, id)
+    }
 }
 
 /// An open [`StdFile`] that belongs to a [`StdFileManager`].
@@ -238,18 +280,16 @@ impl OpenableFile<StdFile> for OpenStdFile {
         self.file.as_ref().and_then(StdFile::id)
     }
 
-    fn execute<W: FileOp<StdFile>>(&mut self, writer: W) -> W::Output {
-        writer.execute(self.file.as_mut().unwrap())
-    }
-
     fn replace_with<C: FnOnce(u64)>(
         self,
-        path: &Path,
+        replacement: StdFile,
         manager: &StdFileManager,
         publish_callback: C,
     ) -> Result<Self, Error> {
         let current_path = self.file.as_ref().unwrap().path.clone();
         self.close()?;
+        let path = replacement.path.clone();
+        replacement.close()?;
 
         std::fs::rename(path, &current_path)?;
         manager.close_handles(&current_path, publish_callback);
@@ -262,6 +302,12 @@ impl OpenableFile<StdFile> for OpenStdFile {
     }
 }
 
+impl OperableFile<StdFile> for OpenStdFile {
+    fn execute<Output, Op: FileOp<Output>>(&mut self, operator: Op) -> Output {
+        operator.execute(self.file.as_mut().unwrap())
+    }
+}
+
 impl Drop for OpenStdFile {
     fn drop(&mut self) {
         if let Some(manager) = &self.manager {
