@@ -5,7 +5,7 @@ use std::{
     convert::Infallible,
     fmt::{Debug, Display},
     fs,
-    ops::RangeBounds,
+    ops::{Deref, DerefMut, RangeBounds},
     path::{Path, PathBuf},
     sync::{
         atomic::{AtomicU16, Ordering},
@@ -15,13 +15,13 @@ use std::{
 
 use flume::Sender;
 use once_cell::sync::Lazy;
-use parking_lot::Mutex;
+use parking_lot::{MappedMutexGuard, Mutex, MutexGuard};
 
 use crate::{
     context::Context,
     error::Error,
     io::{fs::StdFileManager, FileManager, ManagedFile},
-    transaction::{LogEntry, TransactionHandle, TransactionManager},
+    transaction::{LogEntry, ManagedTransaction, TransactionManager},
     tree::{
         self, root::AnyTreeRoot, state::AnyTreeState, EmbeddedIndex, KeyEvaluation, KeySequence,
         Modification, Operation, Reducer, State, TransactableCompaction, TreeFile, TreeRoot,
@@ -81,7 +81,6 @@ impl<File: ManagedFile> Roots<File> {
     }
 
     /// Returns the vault used to encrypt this database.
-    #[must_use]
     pub fn context(&self) -> &Context<File::Manager> {
         &self.data.context
     }
@@ -198,20 +197,21 @@ impl<File: ManagedFile> Roots<File> {
             .iter()
             .zip(states.into_iter())
             .map(|(tree, state)| {
-                tree.borrow().begin_transaction(
-                    transaction.id,
-                    &self.tree_path(tree.borrow().name()),
-                    state.as_ref(),
-                    self.context(),
-                    Some(&self.data.transactions),
-                )
+                tree.borrow()
+                    .begin_transaction(
+                        transaction.id,
+                        &self.tree_path(tree.borrow().name()),
+                        state.as_ref(),
+                        self.context(),
+                        Some(&self.data.transactions),
+                    )
+                    .map(UnlockedTransactionTree::new)
             })
             .collect::<Result<Vec<_>, Error>>()?;
         Ok(ExecutingTransaction {
             roots: self.clone(),
             transaction: Some(transaction),
             trees,
-            transaction_manager: self.data.transactions.clone(),
         })
     }
 }
@@ -241,9 +241,55 @@ impl<File: ManagedFile> Clone for Roots<File> {
 #[must_use]
 pub struct ExecutingTransaction<File: ManagedFile> {
     roots: Roots<File>,
-    transaction_manager: TransactionManager<File::Manager>,
-    trees: Vec<Box<dyn AnyTransactionTree<File>>>,
-    transaction: Option<TransactionHandle>,
+    trees: Vec<UnlockedTransactionTree<File>>,
+    transaction: Option<ManagedTransaction<File::Manager>>,
+}
+
+/// A tree that belongs to an [`ExecutingTransaction`].
+#[must_use]
+pub struct UnlockedTransactionTree<File: ManagedFile>(Mutex<Box<dyn AnyTransactionTree<File>>>);
+
+impl<File: ManagedFile> UnlockedTransactionTree<File> {
+    fn new(file: Box<dyn AnyTransactionTree<File>>) -> Self {
+        Self(Mutex::new(file))
+    }
+
+    /// Locks this tree so that operations can be performed against it.
+    ///
+    /// # Panics
+    ///
+    /// This function panics if `Root` does not match the type specified when
+    /// starting the transaction.
+    pub fn lock<Root: tree::Root>(&self) -> LockedTransactionTree<'_, Root, File> {
+        LockedTransactionTree(MutexGuard::map(self.0.lock(), |tree| {
+            tree.as_mut().as_any_mut().downcast_mut().unwrap()
+        }))
+    }
+}
+
+/// A locked transaction tree. This transactional tree is exclusively available
+/// for writing and reading to the thread that locks it.
+#[must_use]
+pub struct LockedTransactionTree<'transaction, Root: tree::Root, File: ManagedFile>(
+    MappedMutexGuard<'transaction, TransactionTree<Root, File>>,
+);
+
+impl<'transaction, Root: tree::Root, File: ManagedFile> Deref
+    for LockedTransactionTree<'transaction, Root, File>
+{
+    type Target = TransactionTree<Root, File>;
+
+    fn deref(&self) -> &Self::Target {
+        &*self.0
+    }
+}
+
+impl<'transaction, Root: tree::Root, File: ManagedFile> DerefMut
+    for LockedTransactionTree<'transaction, Root, File>
+{
+    fn deref_mut(&mut self) -> &mut Self::Target {
+        &mut *self.0
+    }
 }
 
 impl<File: ManagedFile> ExecutingTransaction<File> {
@@ -251,14 +297,20 @@ impl<File: ManagedFile> ExecutingTransaction<File> {
     #[must_use]
     #[allow(clippy::missing_panics_doc)]
     pub fn entry(&self) -> &LogEntry<'static> {
-        self.transaction.as_ref().unwrap()
+        self.transaction
+            .as_ref()
+            .and_then(|tx| tx.transaction.as_ref())
+            .unwrap()
     }
 
     /// Returns a mutable reference to the [`LogEntry`] for this transaction.
     #[must_use]
     #[allow(clippy::missing_panics_doc)]
     pub fn entry_mut(&mut self) -> &mut LogEntry<'static> {
-        self.transaction.as_mut().unwrap()
+        self.transaction
+            .as_mut()
+            .and_then(|tx| tx.transaction.as_mut())
+            .unwrap()
     }
 
     /// Commits the transaction. Once this function has returned, all data
@@ -271,9 +323,8 @@ impl<File: ManagedFile> ExecutingTransaction<File> {
         let trees = self.roots.data.thread_pool.commit_trees(trees)?;
 
         // Push the transaction to the log.
-        let tree_locks = self
-            .transaction_manager
-            .push(self.transaction.take().unwrap())?;
+        let transaction = self.transaction.take().unwrap();
+        let tree_locks = transaction.commit()?;
 
         // Publish the tree states, now that the transaction has been fully recorded
         for tree in trees {
@@ -286,18 +337,31 @@ impl<File: ManagedFile> ExecutingTransaction<File> {
         Ok(())
     }
 
+    /// Rolls the transaction back. It is not necessary to call this function --
+    /// transactions will automatically be rolled back when the transaction is
+    /// dropped, if `commit()` isn't called first.
+    pub fn rollback(self) {
+        drop(self);
+    }
+
     /// Accesses a locked tree.
     pub fn tree<Root: tree::Root>(
-        &mut self,
+        &self,
         index: usize,
-    ) -> Option<&mut TransactionTree<Root, File>> {
-        self.trees
-            .get_mut(index)
-            .and_then(|any_tree| any_tree.as_mut().as_any_mut().downcast_mut())
+    ) -> Option<LockedTransactionTree<'_, Root, File>> {
+        self.unlocked_tree(index).map(UnlockedTransactionTree::lock)
+    }
+
+    /// Accesses an unlocked tree. Note: If you clone an
+    /// [`UnlockedTransactionTree`], you must make sure to drop all instances
+    /// before calling commit.
+    pub fn unlocked_tree(&self, index: usize) -> Option<&UnlockedTransactionTree<File>> {
+        self.trees.get(index)
     }
 
     fn rollback_tree_states(&mut self) {
         for tree in self.trees.drain(..) {
+            let tree = tree.0.lock();
             tree.rollback();
         }
     }
@@ -420,10 +484,14 @@ impl<Root: tree::Root, File: ManagedFile> TransactionTree<Root, File> {
 
     /// Retrieves the values of `keys`. If any keys are not found, they will be
     /// omitted from the results. Keys are required to be pre-sorted.
-    pub fn get_multiple(
+    pub fn get_multiple<'keys, KeysIntoIter, KeysIter>(
         &mut self,
-        keys: &[&[u8]],
-    ) -> Result<Vec<(ArcBytes<'static>, ArcBytes<'static>)>, Error> {
+        keys: KeysIntoIter,
+    ) -> Result<Vec<(ArcBytes<'static>, ArcBytes<'static>)>, Error>
+    where
+        KeysIntoIter: IntoIterator<Item = &'keys [u8], IntoIter = KeysIter>,
+        KeysIter: Iterator<Item = &'keys [u8]> + ExactSizeIterator,
+    {
         self.tree.get_multiple(keys, true)
     }
 
@@ -657,7 +725,7 @@ impl<Root: tree::Root, File: ManagedFile> Tree<Root, File> {
         key: impl Into<ArcBytes<'static>>,
         value: impl Into<ArcBytes<'static>>,
     ) -> Result<(), Error> {
-        let mut transaction = self.begin_transaction()?;
+        let transaction = self.begin_transaction()?;
         transaction.tree::<Root>(0).unwrap().set(key, value)?;
         transaction.commit()
     }
@@ -701,7 +769,7 @@ impl<Root: tree::Root, File: ManagedFile> Tree<Root, File> {
         key: impl Into<ArcBytes<'static>>,
         value: impl Into<ArcBytes<'static>>,
     ) -> Result<Option<ArcBytes<'static>>, Error> {
-        let mut transaction = self.begin_transaction()?;
+        let transaction = self.begin_transaction()?;
         let existing_value = transaction.tree::<Root>(0).unwrap().replace(key, value)?;
         transaction.commit()?;
         Ok(existing_value)
@@ -714,7 +782,7 @@ impl<Root: tree::Root, File: ManagedFile> Tree<Root, File> {
         keys: Vec<ArcBytes<'a>>,
         operation: Operation<'a, ArcBytes<'static>>,
     ) -> Result<(), Error> {
-        let mut transaction = self.begin_transaction()?;
+        let transaction = self.begin_transaction()?;
         transaction
             .tree::<Root>(0)
             .unwrap()
@@ -726,7 +794,7 @@ impl<Root: tree::Root, File: ManagedFile> Tree<Root, File> {
     /// Removes `key` and returns the existing value, if present. This is executed within its own transaction.
     #[allow(clippy::missing_panics_doc)]
     pub fn remove(&self, key: &[u8]) -> Result<Option<ArcBytes<'static>>, Error> {
-        let mut transaction = self.begin_transaction()?;
+        let transaction = self.begin_transaction()?;
         let existing_value = transaction.tree::<Root>(0).unwrap().remove(key)?;
         transaction.commit()?;
         Ok(existing_value)
@@ -742,7 +810,7 @@ impl<Root: tree::Root, File: ManagedFile> Tree<Root, File> {
         old: Option<&[u8]>,
         new: Option<ArcBytes<'_>>,
     ) -> Result<(), CompareAndSwapError> {
-        let mut transaction = self.begin_transaction()?;
+        let transaction = self.begin_transaction()?;
         transaction
             .tree::<Root>(0)
             .unwrap()
@@ -753,14 +821,18 @@ impl<Root: tree::Root, File: ManagedFile> Tree<Root, File> {
 
     /// Retrieves the values of `keys`. If any keys are not found, they will be
     /// omitted from the results. Keys are required to be pre-sorted.
-    pub fn get_multiple(
+    #[allow(clippy::needless_pass_by_value)]
+    pub fn get_multiple<'keys, Keys>(
         &self,
-        keys: &[&[u8]],
-    ) -> Result<Vec<(ArcBytes<'static>, ArcBytes<'static>)>, Error> {
+        keys: Keys,
+    ) -> Result<Vec<(ArcBytes<'static>, ArcBytes<'static>)>, Error>
+    where
+        Keys: Iterator<Item = &'keys [u8]> + ExactSizeIterator + Clone,
+    {
         catch_compaction_and_retry(|| {
             let mut tree = self.open_for_read()?;
 
-            tree.get_multiple(keys, false)
+            tree.get_multiple(keys.clone(), false)
         })
     }
 
@@ -992,34 +1064,47 @@ where
     sender: flume::Sender<ThreadCommit<File>>,
     receiver: flume::Receiver<ThreadCommit<File>>,
     thread_count: Arc<AtomicU16>,
+    maximum_threads: usize,
 }
 
 impl<File: ManagedFile> ThreadPool<File> {
+    /// Returns a thread pool that will spawn up to `maximum_threads` to process
+    /// file operations.
+    #[must_use]
+    pub fn new(maximum_threads: usize) -> Self {
+        let (sender, receiver) = flume::unbounded();
+        Self {
+            sender,
+            receiver,
+            thread_count: Arc::new(AtomicU16::new(0)),
+            maximum_threads,
+        }
+    }
+
     fn commit_trees(
         &self,
-        mut trees: Vec<Box<dyn AnyTransactionTree<File>>>,
+        trees: Vec<UnlockedTransactionTree<File>>,
     ) -> Result<Vec<Box<dyn AnyTransactionTree<File>>>, Error> {
-        static CPU_COUNT: Lazy<usize> = Lazy::new(num_cpus::get);
-
         // If we only have one tree, there's no reason to split IO across
         // threads. If we have multiple trees, we should split even with one
         // cpu: if one thread blocks, the other can continue executing.
         if trees.len() == 1 {
-            trees[0].commit()?;
-            Ok(trees)
+            let mut tree = trees.into_iter().next().unwrap().0.into_inner();
+            tree.commit()?;
+            Ok(vec![tree])
         } else {
             // Push the trees so that any existing threads can begin processing the queue.
             let (completion_sender, completion_receiver) = flume::unbounded();
             let tree_count = trees.len();
             for tree in trees {
                 self.sender.send(ThreadCommit {
-                    tree,
+                    tree: tree.0.into_inner(),
                     completion_sender: completion_sender.clone(),
                 })?;
             }
 
             // Scale the queue if needed.
-            let desired_threads = tree_count.min(*CPU_COUNT);
+            let desired_threads = tree_count.min(self.maximum_threads);
             loop {
                 let thread_count = self.thread_count.load(Ordering::SeqCst);
                 if (thread_count as usize) >= desired_threads {
@@ -1062,18 +1147,15 @@ impl<File: ManagedFile> Clone for ThreadPool<File> {
             sender: self.sender.clone(),
             receiver: self.receiver.clone(),
             thread_count: self.thread_count.clone(),
+            maximum_threads: self.maximum_threads,
         }
     }
 }
 
 impl<File: ManagedFile> Default for ThreadPool<File> {
     fn default() -> Self {
-        let (sender, receiver) = flume::unbounded();
-        Self {
-            sender,
-            receiver,
-            thread_count: Arc::new(AtomicU16::new(0)),
-        }
+        static CPU_COUNT: Lazy<usize> = Lazy::new(num_cpus::get);
+        Self::new(*CPU_COUNT)
     }
 }
 
@@ -1084,7 +1166,8 @@ fn transaction_commit_thread<File: ManagedFile>(receiver: flume::Receiver<Thread
         completion_sender,
     }) = receiver.recv()
     {
-        let result = tree.commit().map(move |_| tree);
+        let result = tree.commit();
+        let result = result.map(move |_| tree);
         drop(completion_sender.send(result));
     }
 }
@@ -1181,7 +1264,7 @@ mod tests {
         tree.set(b"test", b"value").unwrap();
 
         // Begin a transaction
-        let mut transaction = roots.transaction(&[Versioned::tree("test")]).unwrap();
+        let transaction = roots.transaction(&[Versioned::tree("test")]).unwrap();
 
         // Replace the key with a new value.
         transaction
@@ -1222,7 +1305,7 @@ mod tests {
         tree.set(b"test", b"value").unwrap();
 
         // Begin a transaction
-        let mut transaction = roots.transaction(&[Versioned::tree("test")]).unwrap();
+        let transaction = roots.transaction(&[Versioned::tree("test")]).unwrap();
 
         // Replace the key with a new value.
         transaction
@@ -1232,14 +1315,14 @@ mod tests {
             .unwrap();
 
         // Roll the transaction back
-        drop(transaction);
+        transaction.rollback();
 
         // Ensure that the reader still sees the old value
         let result = tree.get(b"test").unwrap().expect("key not found");
         assert_eq!(result, b"value");
 
         // Begin a new transaction
-        let mut transaction = roots.transaction(&[Versioned::tree("test")]).unwrap();
+        let transaction = roots.transaction(&[Versioned::tree("test")]).unwrap();
         // Check that the transaction has the original value
         let result = transaction
             .tree::<Versioned>(0)
@@ -1421,13 +1504,13 @@ mod tests {
                 ErrorKind::ValueTooLarge
             ));
             // Roll the transaction back
-            drop(transaction);
+            transaction.rollback();
         }
         // Ensure that we can still write to the tree.
         {
             let roots = config.open().unwrap();
 
-            let mut transaction = roots.transaction(&[Versioned::tree("test")]).unwrap();
+            let transaction = roots.transaction(&[Versioned::tree("test")]).unwrap();
 
             // Write some data to the tree
             transaction
