@@ -50,7 +50,6 @@
 use std::{
     borrow::Cow,
     collections::HashMap,
-    convert::TryFrom,
     fmt::{Debug, Display},
     hash::BuildHasher,
     io::SeekFrom,
@@ -69,7 +68,7 @@ use crate::{
     error::Error,
     io::{File, FileManager, FileOp, ManagedFile, ManagedFileOpener, OpenableFile, OperableFile},
     roots::AbortError,
-    transaction::{TransactionHandle, TransactionManager},
+    transaction::{ManagedTransaction, TransactionManager},
     tree::{btree_entry::ScanArgs, serialization::BinarySerialization},
     vault::AnyVault,
     ArcBytes, ChunkCache, CompareAndSwapError, Context, ErrorKind,
@@ -271,7 +270,10 @@ impl<Root: root::Root, File: ManagedFile> TreeFile<Root, File> {
                     let root = Root::deserialize(contents)
                         .map_err(|err| ErrorKind::DataIntegrity(Box::new(err)))?;
                     if let Some(transaction_manager) = transaction_manager {
-                        if !transaction_manager.transaction_was_successful(root.transaction_id())? {
+                        if root.transaction_id() > 0
+                            && !transaction_manager
+                                .transaction_was_successful(root.transaction_id())?
+                        {
                             // The transaction wasn't written successfully, so
                             // we cannot trust the data present.
                             if block_start == 0 {
@@ -442,19 +444,24 @@ impl<Root: root::Root, File: ManagedFile> TreeFile<Root, File> {
     /// Gets the values stored in `keys`. Does not error if a key is missing.
     /// Returns key/value pairs in an unspecified order. Keys are required to be
     /// pre-sorted.
-    #[cfg_attr(feature = "tracing", tracing::instrument(skip(self)))]
-    pub fn get_multiple(
+    #[cfg_attr(feature = "tracing", tracing::instrument(skip(self, keys)))]
+    pub fn get_multiple<'keys, KeysIntoIter, KeysIter>(
         &mut self,
-        keys: &[&[u8]],
+        keys: KeysIntoIter,
         in_transaction: bool,
-    ) -> Result<Vec<(ArcBytes<'static>, ArcBytes<'static>)>, Error> {
+    ) -> Result<Vec<(ArcBytes<'static>, ArcBytes<'static>)>, Error>
+    where
+        KeysIntoIter: IntoIterator<Item = &'keys [u8], IntoIter = KeysIter>,
+        KeysIter: Iterator<Item = &'keys [u8]> + ExactSizeIterator,
+    {
+        let keys = keys.into_iter();
         let mut buffers = Vec::with_capacity(keys.len());
         self.file.execute(TreeGetter {
             from_transaction: in_transaction,
             state: &self.state,
             vault: self.vault.as_deref(),
             cache: self.cache.as_ref(),
-            keys: KeyRange::new(keys.iter().copied()),
+            keys: KeyRange::new(keys),
             key_reader: |key, value| {
                 buffers.push((key, value));
                 Ok(())
@@ -733,7 +740,9 @@ struct TreeCompactor<'a, Root: root::Root, Manager: FileManager> {
     transactions: Option<TransactableCompaction<'a, Manager>>,
     scratch: &'a mut Vec<u8>,
 }
-impl<'a, Root, Manager> FileOp<Result<(Manager::File, TreeCompactionFinisher<'a, Root>), Error>>
+
+impl<'a, Root, Manager>
+    FileOp<Result<(Manager::File, TreeCompactionFinisher<'a, Root, Manager>), Error>>
     for TreeCompactor<'a, Root, Manager>
 where
     Root: root::Root,
@@ -742,7 +751,7 @@ where
     fn execute(
         self,
         file: &mut dyn File,
-    ) -> Result<(Manager::File, TreeCompactionFinisher<'a, Root>), Error> {
+    ) -> Result<(Manager::File, TreeCompactionFinisher<'a, Root, Manager>), Error> {
         let current_path = file.path().to_path_buf();
         let file_name = current_path
             .file_name()
@@ -797,13 +806,13 @@ where
     }
 }
 
-struct TreeCompactionFinisher<'a, Root: root::Root> {
+struct TreeCompactionFinisher<'a, Root: root::Root, Manager: FileManager> {
     state: &'a State<Root>,
     write_state: MutexGuard<'a, ActiveState<Root>>,
-    _transaction: Option<TransactionHandle>,
+    _transaction: Option<ManagedTransaction<Manager>>,
 }
 
-impl<'a, Root: root::Root> TreeCompactionFinisher<'a, Root> {
+impl<'a, Root: root::Root, Manager: FileManager> TreeCompactionFinisher<'a, Root, Manager> {
     fn finish(mut self, new_file_id: u64) {
         self.write_state.file_id = Some(new_file_id);
         self.write_state.publish(self.state);
@@ -1879,18 +1888,12 @@ mod tests {
 
     #[test]
     fn spam_insert_std_versioned() {
-        spam_insert::<Versioned, StdFile>("std");
+        spam_insert::<Versioned, StdFile>("std-versioned");
     }
 
     #[test]
     fn spam_insert_std_unversioned() {
-        spam_insert::<Unversioned, StdFile>("std");
-    }
-
-    #[test]
-    #[cfg(feature = "uring")]
-    fn spam_insert_uring() {
-        spam_insert::<crate::UringFile>("uring");
+        spam_insert::<Unversioned, StdFile>("std-unversioned");
     }
 
     fn spam_insert<R: Root, F: ManagedFile>(name: &str) {
@@ -2007,10 +2010,7 @@ mod tests {
 
         // Get them all
         let mut all_records = tree
-            .get_multiple(
-                &ids.iter().map(ArcBytes::as_slice).collect::<Vec<_>>(),
-                false,
-            )
+            .get_multiple(ids.iter().map(ArcBytes::as_slice), false)
             .unwrap();
         // Order isn't guaranteeed.
         all_records.sort();
